const fs = require('fs');
const fse = require('fs-extra');
const request = require('request');
const path = require('path');
const _ = require('lodash');

const log = require('./log');
const progress = require('./progress');

const baseReqOpts = {
    headers: {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36'
    }
};

const download = (key, fileName, dir, url, proxy) => {
    log.warn(`${fileName} is waiting for download...`);
    let opts = {
        url: url,
        proxy: proxy
    };
    Object.assign(opts, baseReqOpts);
    const maxChunkLen = 20 * 1024 * 1024;
    return new Promise((resolve, reject) => {
        const filePath = path.join(dir, fileName);
        if (fse.existsSync(filePath) || fse.existsSync(`${filePath}.cache`)) {
            resolve(`${filePath} already exists or is downloading.`);
        } else {
            request.get(opts)
                .on('response', response => {
                    const contentLength = response.headers['content-length'];
                    const chunkCount = parseInt(contentLength / maxChunkLen);
                    const lastChunkLength = parseInt(contentLength % maxChunkLen);
                    const onProgress = showProgress(contentLength - 1);
                    if (contentLength > maxChunkLen) {
                        log.info(`the file is big, need to split it to ${lastChunkLength > 0 ? chunkCount + 1 : chunkCount} pieces`);
                        const cache = `${filePath}.cache`;
                        fs.closeSync(fs.openSync(cache, 'w'));
                        const pms = [];
                        const chunks = [];
                        for (let i = 0; i < chunkCount; i++) {
                            const copyOpts = _.cloneDeep(opts);
                            const chunkName = path.join(dir, `${key}_${i}.chunk`);
                            copyOpts.headers['Range'] = `bytes=${i * maxChunkLen}-${(i + 1) * maxChunkLen - 1}`;
                            copyOpts.headers['Connection'] = 'keep-alive';
                            pms.push(downloadChunk(chunkName, copyOpts, onProgress));
                            chunks.push(chunkName);
                        }
                        if (lastChunkLength > 0) {
                            const copyOpts = _.cloneDeep(opts);
                            const chunkName = path.join(dir, `${key}_${chunkCount}.chunk`);
                            copyOpts.headers['Range'] = `bytes=${chunkCount * maxChunkLen}-${contentLength - 1}`;
                            copyOpts.headers['Connection'] = 'keep-alive';
                            pms.push(downloadChunk(chunkName, copyOpts, onProgress));
                            chunks.push(chunkName);
                        }
                        Promise.all(pms).then(msgs => {
                            progress.done();
                            log.info(`${new Date().toLocaleString()}: all pieces have been downloaded! Concat pieces...`);
                            const ws = fse.createWriteStream(filePath, { flag: 'a' });
                            chunks.forEach(chunk => {
                                const bf = fse.readFileSync(chunk);
                                ws.write(bf);
                            });
                            ws.end();
                            chunks.forEach(chunk => {
                                fse.unlinkSync(chunk);
                            });
                            fse.unlinkSync(cache);
                            resolve(`${fileName} has been downloaded!`);
                        }, errs => reject(errs))
                    } else {
                        const copyOpts = _.cloneDeep(opts);
                        copyOpts.headers['Range'] = `bytes=0-${contentLength - 1}`;
                        copyOpts.headers['Connection'] = 'keep-alive';
                        downloadChunk(filePath, copyOpts, onProgress).then(
                            msg => {
                                resolve(msg);
                                progress.done();
                            },
                            err => reject(err)
                        );
                    }
                });
        }
    })
}


const downloadChunk = (filePath, opts, onProcess) => {
    return new Promise((resolve, reject) => {
        log.input(`${new Date().toLocaleString()}: downloading ${filePath}`);
        request.get(opts)
            .on('error', err => {
                reject(err);
            })
            .on('response', response => {
                const ws = fse.createWriteStream(filePath, { encoding: 'binary' });
                response.on('error', err => {
                    return reject(err);
                });
                response.on('data', chunk => {
                    ws.write(chunk);
                    onProcess(chunk.length);
                });
                response.on('end', () => {
                    ws.end();
                    return resolve(`${filePath} has been downloaded!`);
                });
            });
    });
}

const showProgress = (totalSize) => {
    const total = totalSize;
    let current = 0;
    return (size) => {
        current += size;
        progress.showPer(current / total);
    }
}

module.exports = {
    download
};